#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define BUFSIZE 0x1000
#define BUFMIN1 0xFFF
#define MINLEN 3

void help(void);
void list(char *infile);
void extract(char *infile, char *outfile, int index);
void decompress(char *infile, char *outfile);
void compress(char *infile, char *outfile);

void main(int argn, char **argc) {

	switch(argn) {
		case 2:	list(argc[1]);
					break;
		case 3:	extract(argc[1], argc[2], atoi(argc[2]));
					break;
		case 4:	extract(argc[1], argc[2], atoi(argc[3]));
					break;
		default: help();
	}
}

void help(void) {
	printf("Usage:\n\n");
	printf("\tpeace infile\n");
	printf("\tpeace infile index\n");
	printf("\tpeace infile outfile index\n");
	exit(1);
}

void list(char *infile) {

	long length, temp, *buf;
	char type[5];
	FILE *inf;
	int x;

	if(!(inf = fopen(infile, "rb"))) {
		printf("Error opening %s, aborting.\n\n", infile);
		exit(1);
	}

	fseek(inf, 8, 0);
	fread(&length, 4, 1, inf);
	length -= 8;
	type[4] = 0;

	(void *)buf = malloc(length);

	buf[0] = length + 8;
	length = length >> 2;

	for(x=1; x < length; x++)
		fread(&buf[x], 4, 1, inf);

	for(x=0; x < length; x++) {

		if (buf[x+1] > buf[x] + 4) {
			fseek(inf, buf[x], 0);
			fread(&temp, 4, 1, inf);
			if(temp & 0xff000000L) fseek(inf, 1, 1);
			fread(&type, 4, 1, inf);
		}
		else {
			temp = buf[x+1] - buf[x];
			type[0] = 0;
		}

		printf("%d:\t%ld\t%ld\t", x, buf[x], temp & 0xffffffL);

		if(!strcmp(type, "TYPE")) printf("PUD");
		else if(!strcmp(type, "FONT")) printf("FONT");
		else if(!strcmp(type, "FORM")) printf("FORM");
		else if(!strcmp(type, "RIFF")) printf("RIFF");
		else if(!strcmp(type, "SMK2")) printf("SMK2");
		else if((type[0] == 'I') && (type[1] == 'F')) printf("IFF");
		else if((type[0] == 'B') && (type[1] == 'M')) printf("BMP");

		if(temp & 0xff000000L) printf("\t(compressed)");
		printf("\n");
	}

	fclose(inf);
}

void extract(char *infile, char *outfile, int index) {

	long got, length, temp;
	char Buf[BUFSIZE], comp = 0;
	FILE *inf, *outf;

	if(!(inf = fopen(infile, "rb"))) {
		printf("Error opening %s, aborting.\n\n", infile);
		exit(1);
	}

	fseek(inf, 8, 0);
	fread(&length, 4, 1, inf);

	index = (index << 2) + 8;

	if(index >= length) {
		printf("Index greater than number of items in file.\n\n");
		exit(1);
	}

	fseek(inf, index, 0);
	fread(&length, 4, 1, inf);
	fread(&temp, 4, 1, inf);
	fseek(inf, length, 0);
	length = temp - length - 4;

	fread(&got, 4, 1, inf);
	if(got & 0xff000000L) comp = 1;

	if(comp) outf = fopen("temp.$$$", "wb");
	else outf = fopen(outfile, "wb");

	if(!outf) {
		printf("Error writing %s, aborting.\n\n", outfile);
		exit(1);
	}

	got = fread(Buf, 1, BUFSIZE, inf);
	while((got < length) && (!(got & BUFMIN1))) {
		fwrite(Buf, BUFSIZE, 1, outf);
		got += fread(Buf, 1, BUFSIZE, inf);
	}

	if (got & BUFMIN1) fwrite(Buf, got & BUFMIN1, 1, outf);
	else fwrite(Buf, length - got + BUFSIZE, 1, outf);

	fclose(outf);
	fclose(inf);

	if(comp) {
		decompress("temp.$$$", outfile);
		printf("Decompressing...\n");
		unlink("temp.$$$");
	}
}

void decompress(char *infile, char *outfile) {

	unsigned int flags, b1pos, b2pos, got, len, where, i;
	unsigned char Buf1[BUFSIZE], Buf2[BUFSIZE];
	FILE *inf, *outf;

	if(!(inf = fopen(infile, "rb"))) {
		printf("Error opening %s, aborting.\n\n", infile);
		exit(1);
	}

	if(!(outf = fopen(outfile, "wb"))) {
		printf("Error writing %s, aborting.\n\n", outfile);
		exit(1);
	}

	/* clear Buf1, buffer positions, and flags */

	for(i=0; i < BUFSIZE; i++) Buf1[i] = 0;
	b1pos = b2pos = flags = 0;

	/* start the input cache */

	got = fread(Buf2, 1, BUFSIZE, inf);

	/* start decompressing */

	while(b2pos != got) {

		/* read the next flag */

		flags = (flags >> 1);

		if(!(flags & 0x100)) {
			flags = (0xFF00 | Buf2[b2pos]);
			b2pos = ((++b2pos) & BUFMIN1);
			if(!b2pos) got = fread(Buf2, 1, BUFSIZE, inf);
		}

		/* check for EOF */

		if(b2pos == got) continue;

		/* uncompressed byte */

		if((flags & 1) == 1) {
			Buf1[b1pos] = Buf2[b2pos];
			b1pos = ((++b1pos) & BUFMIN1);
			b2pos = ((++b2pos) & BUFMIN1);
			if(!b1pos) fwrite(Buf1, BUFSIZE, 1, outf);
			if(!b2pos) got = fread(Buf2, 1, BUFSIZE, inf);
		}

		/* pos/len pair */

		else {
			where = Buf2[b2pos];
			b2pos = ((++b2pos) & BUFMIN1);
			if(!b2pos) got = fread(Buf2, 1, BUFSIZE, inf);

			if(b2pos == got) continue;

			len = Buf2[b2pos];
			b2pos = ((++b2pos) & BUFMIN1);
			if(!b2pos) got = fread(Buf2, 1, BUFSIZE, inf);

			where = (((len & 0x0F) << 8) | where);
			len = ((len & 0xF0) >> 4) + MINLEN + where;

			for(i=where; i < len; i++) {
				Buf1[b1pos] = Buf1[i & BUFMIN1];;
				b1pos = ((++b1pos) & BUFMIN1);
				if(!b1pos) fwrite(Buf1, BUFSIZE, 1, outf);
			}
		}
	}

	/* flush the output cache */

	fwrite(Buf1, b1pos, 1, outf);

	fclose(inf);
	fclose(outf);
}
